package se.sogeti.jd.crazysnake.shakeasnake;

import se.citerus.crazysnake.Square;

public final class Evaluator {

    public static final int WINNING_SCORE = 1000000;

    public static final int LOOSING_SCORE = -1000000;

    public static final int DRAW_SCORE = 0;

    /** The score of each fruit. */
    protected static final int FRUIT_SCORE = 100;

    /** The minimum space required around a new position for it to be considered safe. */
    protected  static final int MIN_SPACE = 300;

    protected final int cols;
    protected final int rows;

    /** The maximum possible distance. */
    protected final int maxDistance;

    protected Square[][] arena;

    /** The distance (number of moves) to my closest fruit. */
    protected int myClosestDistance;

    /** His closest fruit found by {@code findClosestFruit}. */
    protected Pos hisClosestFruit;

    /** The distance (number of moves) to his closest fruit. */
    protected int hisClosestDistance;

    /** Squares visited during space computation. */
    protected int[][] visitedSquares;

    /** Free space found. */
    protected int numberOfSquares;

    public Evaluator(int cols, int rows) {
        this.cols = cols;
        this.rows = rows;
        this.maxDistance = cols + rows;
    }

    /**
     * Evaluates the position that would be if the snakes moved to {@code myPos} and {@code hisPos}.
     */
    public final int evaluate(Square[][] arena, Pos[] fruitPositions, Pos myNewPos, Pos hisNewPos, int myFruits, int hisFruits) {
        this.arena = arena;

        // If a snake hits a snake or a wall, the other wins
        if (arena[myNewPos.x][myNewPos.y] == Arena.OCCUPIED_SQUARE) {
            return LOOSING_SCORE;
        }
        if (arena[hisNewPos.x][hisNewPos.y] == Arena.OCCUPIED_SQUARE) {
            return WINNING_SCORE;
        }

        // Calculate the number of reachable squares from this position
        numberOfSquares = 0;
        visitedSquares = new int[cols][rows];
        findSpace(myNewPos.x, myNewPos.y);

        // If too little space for us, we loose
        if (numberOfSquares < MIN_SPACE) {
            return LOOSING_SCORE;
        }

        // If a snake eats a fruit
        if (FruitArray.contains(fruitPositions, myNewPos)) {
            myFruits++;
        }
        if (FruitArray.contains(fruitPositions, hisNewPos)) {
            hisFruits++;
        }

        // Calculate base score
        int score = (myFruits - hisFruits) * FRUIT_SCORE;

        // Find closest fruits
        findClosestFruit(arena, myNewPos, hisNewPos, fruitPositions);

        // If there is a fruit which is closer to me, and there is an easy path to it
        if (myClosestDistance != Integer.MAX_VALUE) {
	        // The maximum score for closest fruit is 90
	        int myClosestFruitScore = 90 * (maxDistance - myClosestDistance) / maxDistance;
	        score += myClosestFruitScore;
        }

        // If the snakes collide head to head
        if (myNewPos == hisNewPos) {
            // If we are in the lead, colliding is almost as bad as loosing
            if (score > 0) {
                return LOOSING_SCORE;
            }
            // Otherwise it is a draw
            else {
                return DRAW_SCORE;
            }
        }

        return score;
    }

    /**
     * Finds free space, that is, the number of reachable squares, starting from the given square.
     * Each square is visited at most once.
     */
    private final void findSpace(int x, int y) {
        numberOfSquares++;

        // Mark this square as visited
        visitedSquares[x][y] = 1;

        // If we have not found enough space yet
        // and the square to the right has not been checked
        // and the square to the right is empty
        if ((numberOfSquares < MIN_SPACE) && (visitedSquares[x + 1][y] == 0) && (arena[x + 1][y] == Arena.EMPTY_SQUARE)) {
            findSpace(x + 1, y);
        }
        if ((numberOfSquares < MIN_SPACE) && (visitedSquares[x - 1][y] == 0) && (arena[x - 1][y] == Arena.EMPTY_SQUARE)) {
            findSpace(x - 1, y);
        }
        if ((numberOfSquares < MIN_SPACE) && (visitedSquares[x][y + 1] == 0) && (arena[x][y + 1] == Arena.EMPTY_SQUARE)) {
            findSpace(x, y + 1);
        }
        if ((numberOfSquares < MIN_SPACE) && (visitedSquares[x][y - 1] == 0) && (arena[x][y - 1] == Arena.EMPTY_SQUARE)) {
            findSpace(x, y - 1);
        }
    }

    /**
     * Finds the fruits closest to the given positions, ignoring fruits
     * that are exactly at one of the positions.
     */
    protected void findClosestFruit(Square[][] arena, Pos myPos, Pos hisPos, Pos[] fruitPositions) {
        myClosestDistance = Integer.MAX_VALUE;
        hisClosestFruit = null;
        hisClosestDistance = Integer.MAX_VALUE;

        for (int index = 0; index < fruitPositions.length; index++) {
            Pos fruitPosition = fruitPositions[index];
            if (fruitPosition == null) {
                continue;
            }
            int hisDistance = Math.abs(hisPos.x - fruitPosition.x) + Math.abs(hisPos.y - fruitPosition.y);
            if ((hisDistance > 0) && (hisDistance < hisClosestDistance)) {
                hisClosestFruit = fruitPosition;
                hisClosestDistance = hisDistance;
            }
        }

    	int myX = myPos.x;
    	int myY = myPos.y;
        for (int index = 0; index < fruitPositions.length; index++) {
            Pos fruitPosition = fruitPositions[index];
            if (fruitPosition == null) {
                continue;
            }
        	int fruitX = fruitPosition.x;
        	int fruitY = fruitPosition.y;
            int myDistance = Math.abs(myX - fruitX) + Math.abs(myY - fruitY);
            // If my distance is > 0 (that is, we are not eating this fruit now)
            // and my distance is < my closest distance yet
            if ((myDistance > 0) && (myDistance < myClosestDistance)) {
                // If this fruit is not the same as his closest fruit
            	// or it is the same fruit but I am closer to it
            	if ((fruitPosition != hisClosestFruit) || (myDistance < hisClosestDistance)) {
            		// If there is an easy path from me to the fruit, we go for it.
            		// Otherwise, we skip it to avoid getting stuck in ourselves.
            		if (isEasyPathToPosition(arena, myX, myY, fruitX, fruitY)) {
		                myClosestDistance = myDistance;
            		}
            	}
            }
        }
    }

    /**
     * Returns {@code true} if there is an easy path to follow from the start
     * position to the end position, either going first horizontally, and then
     * vertically, or first going vertically, and then horizontally.
     */
	public static final boolean isEasyPathToPosition(Square[][] arena, int startX, int startY, int endX, int endY) {
		int dX = (endX > startX) ? 1 : ((endX < startX) ? -1 : 0);
		int dY = (endY > startY) ? 1 : ((endY < startY) ? -1 : 0);

		// If we come all the way to the end, we're done
		if (moveHorizontally(arena, startX, startY, endX, dX) && moveVertically(arena, endX, startY, endY, dY)) {
			return true;
		}
		if (moveVertically(arena, startX, startY, endY, dY) && moveHorizontally(arena, startX, endY, endX, dX)) {
			return true;
		}

		// We found no easy path
		return false;
	}

	private static final boolean moveHorizontally(Square[][] arena, int x, int y, int endX, int dX) {
		while (x != endX) {
			x += dX;
			if (arena[x][y] == Arena.OCCUPIED_SQUARE) {
				return false;
			}
		}
		return true;
	}

	private static final boolean moveVertically(Square[][] arena, int x, int y, int endY, int dY) {
		while (y != endY) {
			y += dY;
			if (arena[x][y] == Arena.OCCUPIED_SQUARE) {
				return false;
			}
		}
		return true;
	}
}
