package se.sogeti.jd.crazysnake.shakeasnake;

import java.text.NumberFormat;

import se.citerus.crazysnake.Direction;
import se.citerus.crazysnake.Movement;
import se.sogeti.jd.crazysnake.shakeasnake.api.SnakeImpl;
import se.sogeti.jd.crazysnake.shakeasnake.api.SquareImpl;

public final class Finder {

    /** The maximum time allowed in nano seconds. */
    private static final long MAX_TIME = 49000000;

    /** The maximum depth allowed. */
    private static final int MAX_DEPTH = 50;

    private static final Movement[] MOVES = Movement.values();

    private final Evaluator evaluator;

    /**
     * A stack where undo objects are put when making a move, and removed when
     * undoing the move again. Since there is a limit to search depth, the
     * stack is implemented as a preallocated array of undo objects to avoid
     * some garbage collection.
     */
    private UndoInfo[] undoStack;
    private int undoStackPointer;

    /**
     * A stack where score maps are put when going deeper in the search tree,
     * and removed when going up again. The depth is used stack pointer.
     */
    private ScoreMap[] scoreMapStack;

    private SquareImpl[][] arena;
    private SnakeImpl me;
    private SnakeImpl him;
    private int growFrequency;
    private long turnsUntilGrowth;
    private Pos[] fruitPositions;

    /** The number of evaluated moves during one iteration. */
    private long evaluatedMoves;
    /** The total number of evaluated moves. */
    private long totalEvaluatedMoves;

    private final NumberFormat nf = NumberFormat.getIntegerInstance();

    public Finder(int cols, int rows) {
        this.evaluator = new Evaluator(cols, rows);

        nf.setGroupingUsed(true);

        // Initialize score map stack
        scoreMapStack = new ScoreMap[MAX_DEPTH];
        for (int i = 0; i < MAX_DEPTH; i++) {
            scoreMapStack[i] = new ScoreMap();
        }

        // Initialize the undo stack
        undoStack = new UndoInfo[MAX_DEPTH];
        for (int i = 0; i < MAX_DEPTH; i++) {
        	undoStack[i] = new UndoInfo();
        }
        undoStackPointer = 0;
    }

    /**
     * Finds the best move in the given position.
     */
    public Movement findMove(SquareImpl[][] arena, SnakeImpl me, SnakeImpl him, int growFrequency, long turnsUntilGrowth, Pos[] fruitPositions) {
        this.arena = arena;
        this.me = me;
        this.him = him;
        this.growFrequency = growFrequency;
        this.turnsUntilGrowth = turnsUntilGrowth;
        this.fruitPositions = fruitPositions;
        this.totalEvaluatedMoves = 0;

        // The search depth, start at 1
        int depth = 1;

        // The number of new moves that was evaluated in the last iteration,
        // used to estimate the number of new moves in the next iteration
        long movesEvaluatedLastIteration = 0;

        // The best move so far
        MoveScore bestMove = null;

        // Start counting nano seconds
        long startTime = System.nanoTime();
        long usedTime = System.nanoTime() - startTime;

        while (usedTime < MAX_TIME) {
        	evaluatedMoves = 0;
            MoveScore newBestMove = findMove(depth);
            totalEvaluatedMoves += evaluatedMoves;

            // If best move has score LOOSING, throw away this best move and return
            // the previous best move, and hope he does not make his best move
            if (newBestMove.score == Evaluator.LOOSING_SCORE) {
                if (bestMove == null) {
                    bestMove = newBestMove;
                }
                break;
            }

            // If best move has score WINNING, we don't need to search deeper
        	if (newBestMove.score == Evaluator.WINNING_SCORE) {
                bestMove = newBestMove;
        		break;
        	}
            bestMove = newBestMove;

            // Update used time to time used so far
            usedTime = System.nanoTime() - startTime;
//            Utils.log("Depth " + depth + ", best move is " + bestMove + ", used time is " + nf.format(usedTime) + " ns");
            long remainingTime = MAX_TIME - usedTime;
            // Calculate the number of new moves evaluated this iteration
            long newMovesEvaluatedThisIteration = evaluatedMoves - movesEvaluatedLastIteration;
            // Calculate the time it takes to evaluate one move in ns
            long timeToEvaluateOneMove = usedTime / totalEvaluatedMoves;
//            Utils.log("Evaluated " + evaluatedMoves + " moves (" + newMovesEvaluatedThisIteration + " new) this iteration, " + timeToEvaluateOneMove + " ns/move so far");
            // Calculate the number of moves to evaluate next iteration
            // This is only an estimation, since in some cases, searching
            // will stop earlier with a WINNING or LOOSING score
            long newMovesToEvaluateNextIteration = newMovesEvaluatedThisIteration * 9;
            // Save the number of new moves evaluated this iteration for next time
            movesEvaluatedLastIteration = newMovesEvaluatedThisIteration;
            // Calculate the time it takes to evaluate all moves in next iteration,
            // that is, all the moves evaluated in this iteration plus the estimated
            // number of moves on the next depth
            long estimatedTime = (evaluatedMoves + newMovesToEvaluateNextIteration) * timeToEvaluateOneMove;
//            Utils.log("Estimated new moves next iteration = " + (evaluatedMoves + newMovesToEvaluateNextIteration)
//                    + " (" + newMovesToEvaluateNextIteration + " new), estimated time needed (old + new) = "
//                    + nf.format(estimatedTime) + ", remaining time = " + nf.format(remainingTime));
            // If the estimated time to evaluate all moves in next iteration is less
            // than the remaining time, increase depth, and start next iteration
            if (estimatedTime < remainingTime) {
            	depth++;
            } else {
            	break;
            }
        }

//        long stopTime = System.nanoTime();
//        long movesPerMs = totalEvaluatedMoves * 1000000L / (stopTime - startTime);
//        Utils.log("Depth " + depth + ", best move is " + bestMove + ", Evaluated " + totalEvaluatedMoves +
//                  " moves in " + (stopTime - startTime) + " ns = " + movesPerMs + " moves/ms");

        return bestMove.move;
    }

    /**
     * Finds the best move, searching to the specified depth.
     */
    private MoveScore findMove(int depth) {
        ScoreMap scoreMap = scoreMapStack[depth];

        evaluate(scoreMap);
        if (depth > 1) {
            updateScoreMapWithDeeperScores(scoreMap, depth - 1);
        }
        return scoreMap.getMyBestMove();
    }

    /**
     * Updates the score map by searching deeper, that is, looking more moves ahead,
     * to find a more accurate score.
     */
    private void updateScoreMapWithDeeperScores(ScoreMap scoreMap, int depth) {
        for (Movement myMove : MOVES) {
            for (Movement hisMove : MOVES) {
                int score = scoreMap.getScore(myMove, hisMove);

                // If the existing score is WINNING or LOOSING, we should not look deeper
                if ((score == Evaluator.WINNING_SCORE) || (score == Evaluator.LOOSING_SCORE)) {
                    continue;
                }

                // Make the move that we want to update the score for
                makeMove(myMove, hisMove);
                // Find the best move in the resulting position
                MoveScore bestMove = findMove(depth);
                // The score of the returned best move will become the new
                // score of the move to be updated
                // This score is hopefully more accurate than the old score
                scoreMap.setScore(myMove, hisMove, bestMove.score);
                // Undo the move again
                undoMove();
            }
        }
    }

    /**
     * Evaluates all possible moves in this position, and stores the result in the given score map.
     */
    private void evaluate(ScoreMap scoreMap) {
        for (Movement myMove : MOVES) {
            for (Movement hisMove : MOVES) {
                Pos myOldPos = me.getHeadPosition();
                Pos myNewPos = myOldPos.offset(me.getDirection().turn(myMove));
                Pos hisOldPos = him.getHeadPosition();
                Pos hisNewPos = hisOldPos.offset(him.getDirection().turn(hisMove));
                int score = evaluator.evaluate(arena, fruitPositions, myNewPos, hisNewPos, me.getScore(), him.getScore());
                scoreMap.setScore(myMove, hisMove, score);
                evaluatedMoves++;
            }
        }
    }

    /**
     * Makes the given move(s), updating the snakes and the arena.
     */
    private void makeMove(Movement myMove, Movement hisMove) {
        boolean grow = (turnsUntilGrowth == 0);

        // Calculate new snake positions
        Pos myOldPos = me.getHeadPosition();
        Direction myNewDir = me.getDirection().turn(myMove);
        Pos myNewPos = myOldPos.offset(myNewDir);

        Pos hisOldPos = him.getHeadPosition();
        Direction hisNewDir = him.getDirection().turn(hisMove);
        Pos hisNewPos = hisOldPos.offset(hisNewDir);

        // Save old snake and arena state
        UndoInfo undo = undoStack[undoStackPointer++];

        undo.setMyLastPart(me.getLastPart());
        undo.setMyDirection(me.getDirection());
        undo.setMyScore(me.getScore());
        undo.setHisLastPart(him.getLastPart());
        undo.setHisDirection(him.getDirection());
        undo.setHisScore(him.getScore());
        undo.setTurnsUntilGrowth(turnsUntilGrowth);
        undo.setShrink(grow); // If the snake has grown, it should also shrink
        undo.setMyFruitIndex(FruitArray.remove(fruitPositions, myNewPos));
        undo.setHisFruitIndex(FruitArray.remove(fruitPositions, hisNewPos));

        // Remove old snake parts
        Arena.removeLastSnakePart(arena, me, him, grow);

        // Move snakes
        me.setDirection(myNewDir);
        me.moveTo(myNewPos, grow);
        if (undo.getMyFruitIndex() != -1) {
            me.eat();
        }

        him.setDirection(hisNewDir);
        him.moveTo(hisNewPos, grow);
        if (undo.getHisFruitIndex() != -1) {
            him.eat();
        }

        // Add new snake parts
        Arena.moveSnakeHead(arena, me, him);

        // Decrease the number of turns until growth, since one turn is now over
        turnsUntilGrowth--;
        if (turnsUntilGrowth < 0) {
            turnsUntilGrowth = growFrequency - 1;
        }
    }

    /**
     * Undoes the last move(s), restoring the snakes and arena to their old state.
     */
    private void undoMove() {
    	UndoInfo undo = undoStack[--undoStackPointer];

    	Pos myOldHead = me.getHeadPosition();
    	Pos hisOldHead = him.getHeadPosition();

    	// Move back snakes
    	me.backTo(undo.getMyLastPart(), undo.getShrink());
    	me.setDirection(undo.getMyDirection());
    	me.setScore(undo.getMyScore());
    	him.backTo(undo.getHisLastPart(), undo.getShrink());
    	him.setDirection(undo.getHisDirection());
    	him.setScore(undo.getHisScore());
    	turnsUntilGrowth = undo.getTurnsUntilGrowth();
        int myFruitIndex = undo.getMyFruitIndex();
        int hisFruitIndex = undo.getHisFruitIndex();
        if (myFruitIndex != -1) {
            FruitArray.add(fruitPositions, myOldHead, myFruitIndex);
        }
        if (hisFruitIndex != -1) {
            FruitArray.add(fruitPositions, hisOldHead, hisFruitIndex);
        }

        // Move snake heads backward, including putting any fruits back
    	Arena.backSnakeHead(arena, me, him, myOldHead, hisOldHead);

    	// Add snake tails
    	Arena.addLastSnakePart(arena, me, him);
    }
}
