package se.sogeti.jd.crazysnake.shakeasnake;

import java.util.Set;

import se.citerus.crazysnake.BaseBrain;
import se.citerus.crazysnake.BrainId;
import se.citerus.crazysnake.HeatMeta;
import se.citerus.crazysnake.HeatState;
import se.citerus.crazysnake.Movement;
import se.citerus.crazysnake.Position;
import se.citerus.crazysnake.Snake;
import se.sogeti.jd.crazysnake.shakeasnake.api.SnakeImpl;
import se.sogeti.jd.crazysnake.shakeasnake.api.SquareImpl;

/**
 * The main snake brain class. Maintains an internal state of the snakes and
 * the arena.
 *
 * @author Johan Dykstrom
 */
public final class ShakeASnake extends BaseBrain {

	/** Move memory used to record all moves for later. */
	private StringBuilder memory;

	/** Move finder used to find the best move in the given position. */
    private Finder finder;

    /** True if internal arena has been setup. */
    private boolean isSetup;

    /** Internal arena and snake state. */
    private SquareImpl[][] arena;
    private SnakeImpl internalMe;
    private SnakeImpl internalHim;

	private BrainId myId;
	private BrainId hisId;

    @Override
    public String toString() {
        return getName() + "{" + "memory=" + memory + "}";
    }

    @Override
    public void init(Set<BrainId> participants, HeatMeta meta) {
        super.init(participants, meta);

        memory = new StringBuilder();

        // Initialize arena
        int cols = (int) meta.getBoardDimensions().getWidth();
        int rows = (int) meta.getBoardDimensions().getHeight();
        arena = Arena.create(cols, rows);

        // Initialize static matrix of positions
        Pos.init(cols, rows);

        internalMe = new SnakeImpl();
        internalHim = new SnakeImpl();

        // Store the IDs of the two snakes
        myId = new BrainId(this);
    	for (BrainId id : participants) {
			if(!id.equals(myId)) {
				hisId = id;
			}
		}

    	// Initialize move finder
    	finder = new Finder(cols, rows);

        // Internal state has not yet been setup with snake and fruit positions
        isSetup = false;
    }

    @Override
    public Movement getNextMove(HeatState state) {
        if (!isSetup) {
            setup(state);
        }
        update(state);

        Movement movement;
        try {
            Pos[] fruitPositions = FruitArray.create(state.getFruitPositions());
            movement = finder.findMove(arena, internalMe, internalHim, meta.getGrowFrequency(), state.getTurnsUntilGrowth(), fruitPositions);
        } catch (Exception e) {
            e.printStackTrace();
            movement = Movement.FORWARD;
        }
        memory.append(movement.name().charAt(0));
        return movement;
    }

    /**
     * Sets up internal heat state.
     */
    private void setup(HeatState state) {
        Snake externalMe = state.getSnake(myId);
        Snake externalHim = state.getSnake(hisId);

        // Copy snake state
        Position myPos = externalMe.getHeadPosition();
        Position hisPos = externalHim.getHeadPosition();
        internalMe.setHeadPosition(Pos.valueOf(myPos.getX(), myPos.getY()));
        internalHim.setHeadPosition(Pos.valueOf(hisPos.getX(), hisPos.getY()));

        internalMe.setDirection(externalMe.getDirection());
        internalHim.setDirection(externalHim.getDirection());

        internalMe.setScore(externalMe.getScore());
        internalHim.setScore(externalHim.getScore());

        // Add head part in new head position
        Arena.moveSnakeHead(arena, internalMe, internalHim);

        // Internal state has been setup done
        isSetup = true;
    }

    /**
     * Updates internal heat state.
     */
    private void update(HeatState state) {
        Snake externalMe = state.getSnake(myId);
        Snake externalHim = state.getSnake(hisId);

        // Remove the last part of the snake from the arena
        Arena.removeLastSnakePart(arena, internalMe, internalHim, internalMe.getLength() < externalMe.getLength());

        // Copy snake state
        Position myPos = externalMe.getHeadPosition();
        Position hisPos = externalHim.getHeadPosition();
        internalMe.moveTo(Pos.valueOf(myPos.getX(), myPos.getY()), internalMe.getLength() < externalMe.getLength());
        internalHim.moveTo(Pos.valueOf(hisPos.getX(), hisPos.getY()), internalHim.getLength() < externalHim.getLength());

        internalMe.setDirection(externalMe.getDirection());
        internalHim.setDirection(externalHim.getDirection());

        internalMe.setScore(externalMe.getScore());
        internalHim.setScore(externalHim.getScore());

        // Add head part in new head position
        Arena.moveSnakeHead(arena, internalMe, internalHim);
    }
}
